# -*- coding: utf-8 -*-
#############################################################################
#  This file is part of libforensic1394.                                    #
#  Copyright (C) 2010  Freddie Witherden <freddie@witherden.org>            #
#                                                                           #
#  libforensic1394 is free software: you can redistribute it and/or modify  #
#  it under the terms of the GNU Lesser General Public License as           #
#  published by the Free Software Foundation, either version 3 of the       #
#  License, or (at your option) any later version.                          #
#                                                                           #
#  libforensic1394 is distributed in the hope that it will be useful,       #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#  GNU Lesser General Public License for more details.                      #
#                                                                           #
#  You should have received a copy of the GNU Lesser General Public         #
#  License along with libforensic1394.  If not, see                         #
#  <http://www.gnu.org/licenses/>.                                          #
#############################################################################

from ctypes import cdll, CFUNCTYPE, POINTER, Structure, c_int, c_size_t, \
                   c_uint64, c_int64, c_uint32, c_uint16, c_void_p, c_char, \
                   c_char_p
from ctypes.util import find_library

from forensic1394.errors import process_result

# Try to find the forensic1394 shared library
loc = find_library("forensic1394")

if loc is None:
    raise ImportError

# Open up the library
lib = cdll.LoadLibrary(loc)

# Opaque bus and device pointers
class busptr(c_void_p):
    pass

class devptr(c_void_p):
    pass

# Wrap the forensic1394_req structure
# C def: struct { uint64_t addr, size_t len, void *buf }
class forensic1394_req(Structure):
    _fields_ = [("addr", c_uint64),
                ("len", c_size_t),
                ("buf", c_void_p)]

# Wrap the forensic1394_device_callback type
# C def: void (*forensic1394_device_callback) (forensic1394_bus *bus,
#                                              forensic1394_dev *dev)
forensic1394_device_callback = CFUNCTYPE(None, busptr, devptr)

# Wrap the alloc function
# C def: forensic1394_bus *forensic1394_alloc(void)
forensic1394_alloc = lib.forensic1394_alloc
forensic1394_alloc.argtypes = []
forensic1394_alloc.restype = busptr

# Wrap the enable_sbp2 function
# C def: forensic1394_result forensic1394_enable_sbp2(forensic1394_bus *bus)
forensic1394_enable_sbp2 = lib.forensic1394_enable_sbp2
forensic1394_enable_sbp2.argtypes = [busptr]
forensic1394_enable_sbp2.restype = c_int
forensic1394_enable_sbp2.errcheck = process_result

# Wrap the get devices function
# C def: forensic1394_dev **forensic1394_get_devices(forensic1394_bus *bus,
#                                                    int *ndev,
#                                                    forensic1394_device_callback ondestroy)
forensic1394_get_devices = lib.forensic1394_get_devices
forensic1394_get_devices.argtypes = [busptr, POINTER(c_int), c_void_p]
forensic1394_get_devices.restype = POINTER(devptr)

# Wrap the destroy function
# C def: void forensic1394_destroy(forensic1394_bus *bus)
forensic1394_destroy = lib.forensic1394_destroy
forensic1394_destroy.argtypes = [busptr]
forensic1394_destroy.restype = None

# Wrap the open device function
# C def: forensic1394_result forensic1394_open_device(forensic1394_dev *dev)
forensic1394_open_device = lib.forensic1394_open_device
forensic1394_open_device.argtypes = [devptr]
forensic1394_open_device.restype = c_int
forensic1394_open_device.errcheck = process_result

# Wrap the close device function
# C def: void forensic1394_close_device(forensic1394_dev *dev)
forensic1394_close_device = lib.forensic1394_close_device
forensic1394_close_device.argtypes = [devptr]
forensic1394_close_device.restype = None

# Wrap the is device open function
# C def: int forensic1394_device_is_open(forensic1394_dev *dev)
forensic1394_is_device_open = lib.forensic1394_is_device_open
forensic1394_is_device_open.argtypes = [devptr]
forensic1394_is_device_open.restype = c_int

# Wrap the read device function
# C def: forensic1394_result forensic1394_read_device(forensic1394_dev *dev,
#                                                     uint64_t addr,
#                                                     size_t len, void *buf)
forensic1394_read_device = lib.forensic1394_read_device
forensic1394_read_device.argtypes = [devptr, c_uint64, c_size_t, c_void_p]
forensic1394_read_device.restype = c_int
forensic1394_read_device.errcheck = process_result

# Wrap the vectorised read device function
# C def: forensic1394_result forensic1394_read_device_v(forensic1394_dev *dev,
#                                                       forensic1394_req *req,
#                                                       size_t nreq)
forensic1394_read_device_v = lib.forensic1394_read_device_v
forensic1394_read_device_v.argtypes = [devptr,
                                       POINTER(forensic1394_req),
                                       c_size_t]
forensic1394_read_device_v.restype = c_int
forensic1394_read_device.errcheck = process_result

# Wrap the write device function
# C def: forensic1394_result forensic1394_write_device(forensic1394_dev *dev,
#                                                      uint64_t addr,
#                                                      size_t len, void *buf)
forensic1394_write_device = lib.forensic1394_write_device
forensic1394_write_device.argtypes = [devptr, c_uint64, c_size_t, POINTER(c_char)]
forensic1394_write_device.restype = c_int
forensic1394_write_device.errcheck = process_result

# Wrap the vectorised write device function
# C def: forensic1394_result forensic1394_write_device_v(forensic1394_dev *dev,
#                                                        forensic1394_req *req,
#                                                        size_t nreq)
forensic1394_write_device_v = lib.forensic1394_write_device_v
forensic1394_write_device_v.argtypes = [devptr,
                                        POINTER(forensic1394_req),
                                        c_size_t]
forensic1394_write_device_v.restype = c_int
forensic1394_write_device_v.errcheck = process_result

# Wrap the device CSR function
# C def: void forensic1394_get_device_csr(forensic1394_dev *dev, uint32_t *rom)
forensic1394_get_device_csr = lib.forensic1394_get_device_csr
forensic1394_get_device_csr.argtypes = [devptr, POINTER(c_uint32)]
forensic1394_get_device_csr.restype = None

# Wrap the device node id function
# C def: uint16_t forensic1394_get_device_nodeid(forensic1394_dev *dev)
forensic1394_get_device_nodeid = lib.forensic1394_get_device_nodeid
forensic1394_get_device_nodeid.argtypes = [devptr]
forensic1394_get_device_nodeid.restype = c_uint16

# Wrap the device guid function
# C def: int64_t forensic1394_get_device_guid(forensic1394_dev *dev)
forensic1394_get_device_guid = lib.forensic1394_get_device_guid
forensic1394_get_device_guid.argtypes = [devptr]
forensic1394_get_device_guid.restype = c_int64

# Wrap the device product name function
# C def: const char *forensic1394_get_device_product_name(forensic1394_dev *dev)
forensic1394_get_device_product_name = lib.forensic1394_get_device_product_name
forensic1394_get_device_product_name.argtypes = [devptr]
forensic1394_get_device_product_name.restype = c_char_p

# Wrap the device product id function
# C def: int forensic1394_get_device_product_name(forensic1394_dev *dev)
forensic1394_get_device_product_id = lib.forensic1394_get_device_product_id
forensic1394_get_device_product_id.argtypes = [devptr]
forensic1394_get_device_product_id.restype = c_int

# Wrap the vendor name function
# C def: const char *forensic1394_get_vendor_product_name(forensic1394_dev *dev)
forensic1394_get_device_vendor_name = lib.forensic1394_get_device_vendor_name
forensic1394_get_device_vendor_name.argtypes = [devptr]
forensic1394_get_device_vendor_name.restype = c_char_p

# Wrap the device vendor id function
# C def: int forensic1394_get_device_vendor_name(forensic1394_dev *dev)
forensic1394_get_device_vendor_id = lib.forensic1394_get_device_vendor_id
forensic1394_get_device_vendor_id.argtypes = [devptr]
forensic1394_get_device_vendor_id.restype = c_int

# Wrap the maximum request size function
# C def: int forensic1394_get_device_request_size(forensic1394_dev *dev);
forensic1394_get_device_request_size  = lib.forensic1394_get_device_request_size
forensic1394_get_device_request_size.argtypes = [devptr]
forensic1394_get_device_request_size.restype = c_int

# Wrap the error string function
# C def: const char *forensic1394_get_result_str(forensic1394_result r);
forensic1394_get_result_str = lib.forensic1394_get_result_str
forensic1394_get_result_str.argtypes = [c_int]
forensic1394_get_result_str.restype = c_char_p



# -*- coding: utf-8 -*-
#############################################################################
#  This file is part of libforensic1394.                                    #
#  Copyright (C) 2010  Freddie Witherden <freddie@witherden.org>            #
#                                                                           #
#  libforensic1394 is free software: you can redistribute it and/or modify  #
#  it under the terms of the GNU Lesser General Public License as           #
#  published by the Free Software Foundation, either version 3 of the       #
#  License, or (at your option) any later version.                          #
#                                                                           #
#  libforensic1394 is distributed in the hope that it will be useful,       #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#  GNU Lesser General Public License for more details.                      #
#                                                                           #
#  You should have received a copy of the GNU Lesser General Public         #
#  License along with libforensic1394.  If not, see                         #
#  <http://www.gnu.org/licenses/>.                                          #
#############################################################################

from ctypes import create_string_buffer, byref, cast, POINTER, \
                   c_char, c_size_t, c_uint32, c_void_p

from forensic1394.errors import process_result, Forensic1394StaleHandle

from forensic1394.functions import forensic1394_open_device, \
                                   forensic1394_close_device, \
                                   forensic1394_is_device_open, \
                                   forensic1394_read_device_v, \
                                   forensic1394_write_device_v, \
                                   forensic1394_get_device_csr, \
                                   forensic1394_get_device_nodeid, \
                                   forensic1394_get_device_guid, \
                                   forensic1394_get_device_product_name, \
                                   forensic1394_get_device_product_id, \
                                   forensic1394_get_device_vendor_name, \
                                   forensic1394_get_device_vendor_id, \
                                   forensic1394_get_device_request_size, \
                                   forensic1394_req

from functools import wraps

def checkStale(f):
    @wraps(f)
    def newf(self, *args, **kwargs):
        if self._stale:
            raise Forensic1394StaleHandle
        return f(self, *args, **kwargs)
    return newf

class Device(object):
    def __init__(self, bus, devptr):
        """
        Constructs a new Device instance.  This should not usually be called
        directly; instead a list of pre-constructed Device instances should be
        requested from the bus.
        """

        # Retain a reference to the Bus (otherwise unused)
        self._bus = bus

        # Set as _as_parameter_ to allow passing of self to functions
        self._as_parameter_ = devptr

        # We are not stale
        self._stale = False

        # Copy over the device properties
        self._nodeid = forensic1394_get_device_nodeid(self)
        self._guid = forensic1394_get_device_guid(self)

        self._product_name = forensic1394_get_device_product_name(self)
        self._product_id = forensic1394_get_device_product_id(self)

        self._vendor_name = forensic1394_get_device_vendor_name(self)
        self._vendor_id = forensic1394_get_device_vendor_id(self)

        self._request_size = forensic1394_get_device_request_size(self)

        self._csr = (c_uint32 * 256)()
        forensic1394_get_device_csr(self, self._csr)

    def __del__(self):
        if self.isopen():
            self.close()

    @checkStale
    def open(self):
        """
        Attempts to open the device.  If the device can not be opened, or if the
        device is stale, an exception is raised.
        """
        forensic1394_open_device(self)

    def close(self):
        """
        Closes the device.  If the device is stale this is a no-op
        """
        if not self._stale:
            forensic1394_close_device(self)

    def isopen(self):
        """
        Checks to see if the device is open or not, returning a boolean value.
        In the case of a stale handle False is returned.
        """
        if self._stale:
            return False
        else:
            return bool(forensic1394_is_device_open(self))

    def _readreq(self, req, buf):
        """
        Internal low level read function.
        """
        assert self.isopen()

        # Get a pointer directly into the buffer which we can perform
        # arithmetic on. Compared to cast(byref(buf, off), c_void_p)
        # directly accessing the pointer gives a ~10% performance
        # improvement for scatter requests.
        pbuf = cast(buf, c_void_p).value

        # Create the request tuples
        init = []
        for addr, numb in req:
            init.append((addr, numb, c_void_p(pbuf)))
            pbuf += numb

        # Create the array of forensic1394 requests
        creq = (forensic1394_req * len(req))(*init)

        # Dispatch the requests
        forensic1394_read_device_v(self, creq, len(creq))

    @checkStale
    def read(self, addr, numb, buf=None):
        """
        Attempts to read numb bytes from the device starting at addr.
        The device must be open and the handle can not be stale.
        Requests larger than self.request_size will automatically be
        broken down into smaller chunks.  The resulting data is
        returned.  An exception is raised should an error occur.  The
        optional buf parameter can be used to pass a specific ctypes
        c_char array to read into.  If no buffer is passed then
        create_string_buffer will be used to allocate one.
        """
        if buf == None:
            # No buffer passed; allocate one
            buf = create_string_buffer(numb)

        # Break the request up into rs size chunks; if numb % rs = 0 then
        # lens may have an extra element; zip will take care of this
        rs = self._request_size
        addrs = range(addr, addr + numb, rs)
        lens = [rs] * (numb // rs) + [numb % rs]

        self._readreq(list(zip(addrs, lens)), buf)

        return buf.raw

    @checkStale
    def readv(self, req):
        """
        Performs a batch of read requests of the form: [(addr1, len1),
        (addr2, len2), ...] and returns a generator yielding, in
        sequence, (addr1, buf1), (addr2, buf2), ..., .  This is useful
        when performing a series of `scatter reads' from a device.
        """
        # Create the request buffer
        buf = create_string_buffer(sum(numb for _addr, numb in req))

        # Use _readreq to read the requests into buf
        self._readreq(req, buf)

        # Generate the resulting buffers
        off = 0
        for addr, numb in req:
            yield (addr, buf.raw[off:off + numb])
            off += numb

    @checkStale
    def write(self, addr, buf):
        """
        Attempts to write len(buf) bytes to the device starting at addr.  The
        device must be open and the handle can not be stale.  Requests larger
        than self.request_size will automatically be broken down into smaller
        chunks.  Uses writev internally.
        """
        # Break up the request
        req = []
        for off in range(0, len(buf), self._request_size):
            req.append((addr + off, buf[off:off + self._request_size]))

        # Dispatch
        self.writev(req)

    @checkStale
    def writev(self, req):
        assert self.isopen()

        # Prepare the request array (addr, len, buf)
        creq = (forensic1394_req * len(req)) \
               (*[(addr, len(buf), cast(buf, c_void_p)) \
                  for addr, buf in req])

        # Send off the requests
        forensic1394_write_device_v(self, creq, len(creq))

    @property
    def nodeid(self):
        """
        The node ID of the device on the bus.
        """
        return self._nodeid

    @property
    def guid(self):
        """
        The 48-bit GUID of the device.
        """
        return self._guid

    @property
    def product_name(self):
        """
        The product name of the device; may be ''.
        """
        return self._product_name

    @property
    def product_id(self):
        """
        The product id of the device; integer.
        """
        return self._product_id

    @property
    def vendor_name(self):
        """
        The vendor name of the device; may be ''.
        """
        return self._vendor_name

    @property
    def vendor_id(self):
        """
        The vendor id of the device; integer.
        """
        return self._vendor_id

    @property
    def request_size(self):
        """
        The maximum request size supported by the device in bytes; this is
        always a power of two.
        """
        return self._request_size

    @property
    def csr(self):
        """
        Configuration status ROM for the device, list of 32-bit host-endian
        integers.
        """
        return self._csr



# -*- coding: utf-8 -*-
#############################################################################
#  This file is part of libforensic1394.                                    #
#  Copyright (C) 2010  Freddie Witherden <freddie@witherden.org>            #
#                                                                           #
#  libforensic1394 is free software: you can redistribute it and/or modify  #
#  it under the terms of the GNU Lesser General Public License as           #
#  published by the Free Software Foundation, either version 3 of the       #
#  License, or (at your option) any later version.                          #
#                                                                           #
#  libforensic1394 is distributed in the hope that it will be useful,       #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#  GNU Lesser General Public License for more details.                      #
#                                                                           #
#  You should have received a copy of the GNU Lesser General Public         #
#  License along with libforensic1394.  If not, see                         #
#  <http://www.gnu.org/licenses/>.                                          #
#############################################################################

class ResultCode(object):
    """
    Possible result codes from a forensic1394 function call.  These are
    extracted from the forensic1394.h file.
    """
    Success     = 0
    OtherError  = -1
    BusReset    = -2
    NoPerm      = -3
    Busy        = -4
    IOError     = -5
    IOSize      = -6
    IOTimeout   = -7

class Forensic1394Exception(Exception):
    pass

class Forensic1394ImportError(Forensic1394Exception, ImportError):
    pass

class Forensic1394BusReset(Forensic1394Exception, IOError):
    pass

class Forensic1394StaleHandle(Forensic1394Exception, IOError):
    pass

def process_result(result, fn, args):
    # Call was successful
    if result == ResultCode.Success:
        return

    # Perform a local import to avoid cyclic dependencies
    from forensic1394.functions import forensic1394_get_result_str

    err = fn.__name__ + ": " + forensic1394_get_result_str(result).decode()

    # Decide which exception to throw
    if result == ResultCode.BusReset:
        raise Forensic1394BusReset(err)
    else:
        raise IOError(err)



# -*- coding: utf-8 -*-
#############################################################################
#  This file is part of libforensic1394.                                    #
#  Copyright (C) 2010  Freddie Witherden <freddie@witherden.org>            #
#                                                                           #
#  libforensic1394 is free software: you can redistribute it and/or modify  #
#  it under the terms of the GNU Lesser General Public License as           #
#  published by the Free Software Foundation, either version 3 of the       #
#  License, or (at your option) any later version.                          #
#                                                                           #
#  libforensic1394 is distributed in the hope that it will be useful,       #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#  GNU Lesser General Public License for more details.                      #
#                                                                           #
#  You should have received a copy of the GNU Lesser General Public         #
#  License along with libforensic1394.  If not, see                         #
#  <http://www.gnu.org/licenses/>.                                          #
#############################################################################

from ctypes import c_int, byref

from forensic1394.errors import process_result

from forensic1394.functions import forensic1394_alloc, forensic1394_destroy, \
                                   forensic1394_enable_sbp2, \
                                   forensic1394_get_devices
from forensic1394.device import Device

import weakref

class Bus(object):
    def __init__(self):
        # Allocate a new bus handle; _as_parameter_ allows passing of self
        self._as_parameter_ = forensic1394_alloc()

        # Weak references to the most recent device list
        self._wrefdev = []

    def enable_sbp2(self):
        # Re-raise for a cleaner stack trace
        forensic1394_enable_sbp2(self)

    def devices(self):
        # Mark any active device handles as being stale
        for wdev in self._wrefdev:
            if wdev():
                wdev()._stale = True

        # Clear the current list of weak references
        self._wrefdev = []

        dev = []
        ndev = c_int(0)

        # Query the list of devices attached to the system
        devlist = forensic1394_get_devices(self, byref(ndev), None)

        # If ndev is < 0 then it contains a result status code
        if ndev.value < 0:
            process_result(ndev.value, forensic1394_get_devices, ())

        # Create Device instances for the devices found
        for i in range(0, ndev.value):
            d = Device(self, devlist[i])
            dev.append(d)
            # Maintain a weak reference to this device
            self._wrefdev.append(weakref.ref(d))

        # Return the device list
        return dev

    def __del__(self):
        forensic1394_destroy(self)
